#!/usr/bin/env python3

import sys
import os
import array
import hashlib

def main():
    # If DRY_RUN=1, this program emits what would be executed in
    # the real run.
    dry_run = False
    if 'DRY_RUN' in os.environ and os.environ['DRY_RUN'] == '1':
        dry_run = True

    if len(sys.argv) <= 2:
        print(f"使い方: {sys.argv[0]} オリジナルファイル IPSファイル")
        print("オリジナルのファイルは変更されません。新しいファイルが作られます。")
        sys.exit(1)
    origfilename = sys.argv[1]
    ipsfilename = sys.argv[2]
    (_, origfileext) = os.path.splitext(origfilename)
    outfilename = os.path.splitext(ipsfilename)[0] + origfileext
    if outfilename == origfilename:
        print('オリジナルとパッチ適用後のファイル名が同じです。')
        sys.exit(2)
    origdata = None
    with open(origfilename, mode='rb') as origf:
        origdata = origf.read()
    md5 = hashlib.md5(origdata).hexdigest()
    print(f'オリジナルのMD5: {md5}')
    sha256 = hashlib.sha256(origdata).hexdigest()
    print(f'オリジナルのSHA256: {sha256}')
    if dry_run:
        print(f'パッチ適用後のデータを{outfilename}に書き込みます。')
    with open(ipsfilename, mode='rb') as f:
        b = f.read()
    out = array.array('B')
    out.frombytes(origdata)
    l = len(b)
    if l <= 5 or b[0:5] != b'PATCH':
        print(f"Error: 正しい IPS ファイルではない: {ipsfilename}")
        sys.exit(1)
    ptr = 5
    while ptr < l:
        if ptr + 3 == l and b[ptr:ptr + 3] == b'EOF':
            break
        # https://w.atwiki.jp/snesspec/pages/170.html
        offset = int.from_bytes(b[ptr:ptr + 3], 'big')
        size = int.from_bytes(b[ptr + 3:ptr + 5], 'big')
        if size == 0:
            # RLE hunk: http://fileformats.archiveteam.org/wiki/IPS_(binary_patch_format)
            size = int.from_bytes(b[ptr + 5:ptr + 7], 'big')
            byte = b[ptr + 7]
            if dry_run:
                # print(f'{hex(offset)} to {hex(offset + size)} (len = {hex(size)}) が {byte} で上書きされる予定です。')
                pass
            if len(out) < offset + size:
                diff = offset + size - len(out)
                out.fromlist([0] * diff)
            for j in range(size):
                out[offset + j] = byte
            ptr += 8
            continue
        if dry_run:
            # print(f'{hex(offset)} to {hex(offset + size)} (len = {hex(size)}) が上書きされる予定です。')
            pass
        else:
            pass
        if len(out) < offset + size:
            diff = offset + size - len(out)
            out.fromlist([0] * diff)
        for j in range(size):
            val = b[ptr + 5 + j]
            out[offset + j] = val
        ptr += 5 + size
    if len(origdata) < len(out):
        print(f'ファイルが{len(origdata)} ({hex(len(origdata))})バイトから{len(out)} ({hex(len(out))})バイトに拡張されます。')
    md5 = hashlib.md5(out.tobytes()).hexdigest()
    print(f'パッチ適用後のMD5: {md5}')
    sha256 = hashlib.sha256(out.tobytes()).hexdigest()
    print(f'パッチ適用後のSHA256: {sha256}')
    if dry_run:
        print(f'ファイル{outfilename}の作成・上書きが行われる予定です。')
    else:
        with open(outfilename, mode='wb') as f:
            f.write(out.tobytes())
        print(f'ファイル{outfilename}の作成・上書きが行われました。')
main()
